package main

import (
	"cmp"
	"image/color"
	"log"
	"math"
	"math/rand"
	"slices"
	"strconv"

	"golang.org/x/image/font"
	"golang.org/x/image/font/opentype"

	"github.com/hajimehoshi/ebiten/v2"
	"github.com/hajimehoshi/ebiten/v2/examples/resources/fonts"
	"github.com/hajimehoshi/ebiten/v2/inpututil"
	"github.com/hajimehoshi/ebiten/v2/text"
)

var (
	tileImage *ebiten.Image
	tileFont  font.Face
)

const (
	fontSize              = 48
	innerTileSize         = 90
	tileSize              = 100
	tileSpeed     float64 = 1

	tilePadding = (tileSize - innerTileSize) / 2
)

type Direction int

const (
	DirectionNone Direction = iota
	DirectionLeft
	DirectionRight
	DirectionUp
	DirectionDown
)

type Tile struct {
	screenPos struct{ x, y float64 }
	value     int
	hasMerged bool
}

type Game struct {
	tiles  [4][4]Tile
	isLoss bool

	pressedKeys []ebiten.Key
}

func init() {
	ttf, err := opentype.Parse(fonts.MPlus1pRegular_ttf)

	if err != nil {
		log.Fatal(err)
	}

	tileFont, err = opentype.NewFace(ttf, &opentype.FaceOptions{
		DPI:     72,
		Size:    fontSize,
		Hinting: font.HintingVertical,
	})

	if err != nil {
		log.Fatal(err)
	}

	tileImage = ebiten.NewImage(innerTileSize, innerTileSize)
}

func (g *Game) updateLoss() {
	copy := *g
	g.isLoss = !(copy.stepTiles(DirectionLeft) ||
		copy.stepTiles(DirectionRight) ||
		copy.stepTiles(DirectionUp) ||
		copy.stepTiles(DirectionDown))
}

func (g *Game) IsFull() bool {
	for _, row := range g.tiles {
		for _, tile := range row {
			if tile.value == 0 {
				return false
			}
		}
	}
	return true
}

func (g *Game) SpawnTile() {
	if g.IsFull() {
		g.isLoss = true
		return
	}

	x := rand.Intn(len(g.tiles))
	y := rand.Intn(len(g.tiles[x]))

	for g.tiles[y][x].value != 0 {
		x = rand.Intn(len(g.tiles))
		y = rand.Intn(len(g.tiles[x]))
	}
	if rand.Intn(10) == 0 {
		g.tiles[y][x].value = 4
	} else {
		g.tiles[y][x].value = 2
	}
}

func (g *Game) stepTiles(direction Direction) (hasChanged bool) {
	switch direction {
	case DirectionLeft:
		for y := range g.tiles {
			for x := 1; x < len(g.tiles[y]); x++ {
				if g.tiles[y][x-1].value == 0 {
					g.tiles[y][x-1] = g.tiles[y][x]
					g.tiles[y][x] = Tile{}
					hasChanged = true
				} else if g.tiles[y][x-1].value == g.tiles[y][x].value && !g.tiles[y][x-1].hasMerged && !g.tiles[y][x].hasMerged {
					g.tiles[y][x-1].value *= 2
					g.tiles[y][x-1].hasMerged = true
					g.tiles[y][x] = Tile{}
					hasChanged = true
				}
			}
		}
	case DirectionRight:
		for y := range g.tiles {
			for x := len(g.tiles[y]) - 2; x >= 0; x-- {
				if g.tiles[y][x+1].value == 0 {
					g.tiles[y][x+1] = g.tiles[y][x]
					g.tiles[y][x] = Tile{}
					hasChanged = true
				} else if g.tiles[y][x+1].value == g.tiles[y][x].value && !g.tiles[y][x+1].hasMerged && !g.tiles[y][x].hasMerged {
					g.tiles[y][x+1].value *= 2
					g.tiles[y][x+1].hasMerged = true
					g.tiles[y][x] = Tile{}
					hasChanged = true
				}
			}
		}
	case DirectionUp:
		for y := 1; y < len(g.tiles); y++ {
			for x := range g.tiles[y] {
				if g.tiles[y-1][x].value == 0 {
					g.tiles[y-1][x] = g.tiles[y][x]
					g.tiles[y][x] = Tile{}
					hasChanged = true
				} else if g.tiles[y-1][x].value == g.tiles[y][x].value && !g.tiles[y-1][x].hasMerged && !g.tiles[y][x].hasMerged {
					g.tiles[y-1][x].value *= 2
					g.tiles[y-1][x].hasMerged = true
					g.tiles[y][x] = Tile{}
					hasChanged = true
				}
			}
		}
	case DirectionDown:
		for y := len(g.tiles) - 2; y >= 0; y-- {
			for x := range g.tiles[y] {
				if g.tiles[y+1][x].value == 0 {
					g.tiles[y+1][x] = g.tiles[y][x]
					g.tiles[y][x] = Tile{}
					hasChanged = true
				} else if g.tiles[y+1][x].value == g.tiles[y][x].value && !g.tiles[y+1][x].hasMerged && !g.tiles[y][x].hasMerged {
					g.tiles[y+1][x].value *= 2
					g.tiles[y+1][x].hasMerged = true
					g.tiles[y][x] = Tile{}
					hasChanged = true
				}
			}
		}
	}

	return
}

func (g *Game) MoveTiles(direction Direction) {
	for y := range g.tiles {
		for x := range g.tiles[y] {
			g.tiles[y][x].hasMerged = false
		}
	}

	for {
		old := g.tiles
		g.stepTiles(direction)
		if slices.Equal(g.tiles[:], old[:]) {
			break
		}
	}
}

func (g *Game) Animate() {
	for y, row := range g.tiles {
		for x := range row {
			var targetX = float64(x*tileSize + tilePadding)
			var targetY = float64(y*tileSize + tilePadding)
			g.tiles[y][x].screenPos.x += float64(cmp.Compare(targetX, g.tiles[y][x].screenPos.x)) * tileSpeed
			g.tiles[y][x].screenPos.y += float64(cmp.Compare(targetY, g.tiles[y][x].screenPos.y)) * tileSpeed
		}
	}
}

func (g *Game) Update() error {
	g.Animate()
	g.pressedKeys = inpututil.AppendJustPressedKeys(g.pressedKeys[:0])

	var direction Direction = DirectionNone

Loop:
	for _, key := range g.pressedKeys {
		switch key {
		case ebiten.KeyLeft:
			direction = DirectionLeft
			break Loop

		case ebiten.KeyRight:
			direction = DirectionRight
			break Loop

		case ebiten.KeyUp:
			direction = DirectionUp
			break Loop

		case ebiten.KeyDown:
			direction = DirectionDown
			break Loop
		}
	}

	if direction != DirectionNone {
		old := g.tiles
		g.MoveTiles(direction)
		if !slices.Equal(g.tiles[:], old[:]) {
			g.SpawnTile()
		}
		g.updateLoss()
	}

	return nil
}

func (g *Game) Draw(screen *ebiten.Image) {
	for _, row := range g.tiles {
		for _, tile := range row {
			drawOptions := ebiten.DrawImageOptions{}
			drawOptions.GeoM.Translate(tile.screenPos.x, tile.screenPos.y)

			tileImage.Fill(color.RGBA{R: uint8(50 * math.Log2(float64(tile.value))), G: 0, B: 0, A: 255})

			if tile.value != 0 {
				str := strconv.Itoa(tile.value)
				textBound := text.BoundString(tileFont, str)

				text.Draw(
					tileImage,
					str,
					tileFont,
					innerTileSize/2-textBound.Dx()/2,
					innerTileSize/2+textBound.Dy()/2,
					color.White,
				)
			}

			screen.DrawImage(tileImage, &drawOptions)
		}
	}

	if g.isLoss {
		str := "You lost"
		textBound := text.BoundString(tileFont, str)
		text.Draw(
			screen,
			str,
			tileFont,
			screen.Bounds().Dx()/2-textBound.Dx()/2,
			screen.Bounds().Dy()/2+textBound.Dy()/2,
			color.White,
		)
	}
}

func (g *Game) Layout(outsideWidth, outsideHeight int) (screenWidth, screenHeight int) {
	return 4 * tileSize, 4 * tileSize
}

func main() {
	ebiten.SetWindowSize(640, 640)
	ebiten.SetWindowTitle("Hello, World!")

	game := Game{}
	game.SpawnTile()

	if err := ebiten.RunGame(&game); err != nil {
		log.Fatal(err)
	}
}
